# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of numerical MD integrators.

These integrators are representations of engines for performing
molecular dynamics. Typically they will propagate Newtons
equations of motion in time numerically.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

MDEngine (:py:class:`.MDEngine`)
    Base class for internal MDEngines.

RandomWalk (:py:class:`.RandomWalk`)
    A Random Walk integrator.

Verlet (:py:class:`.Verlet`)
    A Verlet MD integrator.

VelocityVerlet (:py:class:`.VelocityVerlet`)
    A Velocity Verlet MD integrator.

Langevin (:py:class:`.Langevin`)
    A Langevin MD integrator.
"""
import logging
import numpy as np
from pyretis.core.random_gen import create_random_generator
from pyretis.core.particlefunctions import (calculate_kinetic_energy,
                                            calculate_thermo,
                                            calculate_thermo_path,
                                            reset_momentum)
from pyretis.engines.engine import EngineBase

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['MDEngine', 'RandomWalk', 'Verlet', 'VelocityVerlet', 'Langevin']


class MDEngine(EngineBase):
    """Base class for internal MD integrators.

    This class defines an internal MD integrator. This class of
    integrators work with the positions and velocities of the system
    object directly. Further, we make use of the system object in
    order to update forces etc.

    Attributes
    ----------
    timestep : float
        Time step for the integration.
    description : string
        Description of the MD integrator.
    dynamics : str
        A short string to represent the type of dynamics produced
        by the integrator (NVE, NVT, stochastic, ...).

    """

    engine_type = 'internal'

    def __init__(self, timestep, description, dynamics=None):
        """Set up the integrator.

        Parameters
        ----------
        timestep : float
            The time step for the integrator in internal units.
        description : string
            A short description of the integrator.
        dynamics : string or None, optional
            Description of the kind of dynamics the integrator does.

        """
        super().__init__(description)
        self.timestep = timestep
        self.dynamics = dynamics

    def integration_step(self, _):
        """Perform a single time step of the integration.

        Parameters
        ----------
        - : place holder

        Returns
        -------
        out : None
            Does not return anything, in derived classes it will
            typically update the given `System`.

        """
        raise NotImplementedError

    def select_thermo_function(self, thermo='full'):
        """Select function for calculating thermodynamic properties.

        Parameters
        ----------
        thermo : string, or None, optional
            String which selects the kind of thermodynamic output.

        Returns
        -------
        thermo_func : callable or None
            The function matching the requested thermodynamic output.

        """
        thermo_func = None
        if thermo is not None:
            if thermo not in ('full', 'path'):
                logger.debug(
                    'Unknown thermo "%s" requested: Using "path"!',
                    thermo,
                )
                thermo = 'path'
            logger.debug('Thermo output for %s.integrate(): "%s"',
                         self.__class__.__name__, thermo)
            if thermo == 'full':
                thermo_func = calculate_thermo
            elif thermo == 'path':
                thermo_func = calculate_thermo_path
        return thermo_func

    def integrate(self, ensemble, steps, thermo='full'):
        """Perform several integration steps.

        This method will perform several integration steps, but it will
        also calculate order parameter(s) if requested and energy
        terms.

        Parameters
        ----------
        ensemble: dict
            It contains:

            * `system` : object like :py:class:`.System`
              The system we are integrating.
            * `order_function` : object like :py:class:`.OrderParameter`
              An order function can be specified if we want to
              calculate the order parameter along with the simulation.

        steps : integer
            The number of steps we are going to perform. Note that we
            do not integrate on the first step (e.g. step 0) but we do
            obtain the other properties. This is to output the starting
            configuration.
        thermo : string, optional
            Select the thermodynamic properties we are to calculate.

        Yields
        ------
        results : dict
            The result of a MD step. This contains the state of the
            system and also the order parameter(s) (if calculated) and
            the thermodynamic quantities (if calculated).

        """
        thermo_func = self.select_thermo_function(thermo=thermo)
        system = ensemble['system']
        order_function = ensemble.get('order_function')
        system.potential_and_force()
        for i in range(steps):
            if i == 0:
                pass
            else:
                self.integration_step(system)
            results = {'system': system}
            if order_function is not None:
                results['order'] = self.calculate_order(ensemble)
            if thermo_func:
                results['thermo'] = thermo_func(system)
            yield results

    def invert_dt(self):
        """Invert the time step for the integration.

        Returns
        -------
        out : boolean
            True if the time step is positive, False otherwise.

        """
        self.timestep *= -1.0
        return self.timestep > 0.0

    def propagate(self, path, ensemble, reverse=False):
        """Generate a path by integrating until a criterion is met.

        This function will generate a path by calling the function
        specifying the integration step repeatedly. The integration is
        carried out until the order parameter has passed the specified
        interfaces or if we have integrated for more than a specified
        maximum number of steps. The given system defines the initial
        state and the system is reset to its initial state when this
        method is done.

        Parameters
        ----------
        path : object like :py:class:`.PathBase`
            This is the path we use to fill in phase-space points.
            We are here not returning a new path, this since we want
            to delegate the creation of the path (type) to the method
            that is running `propagate`.
        ensemble: dict
            It contains:

            * `system` : object like :py:class:`.System`
              The system object gives the initial state for the
              integration. The initial state is stored and the system is
              reset to the initial state when the integration is done.
            * `order_function` : object like :py:class:`.OrderParameter`
              An order function can be specified if we want to
              calculate the order parameter along with the simulation.
            * `interfaces` : list of floats
              These interfaces define the stopping criterion.

        reverse : boolean, optional
            If True, the system will be propagated backward in time.

        Returns
        -------
        success : boolean
            This is True if we generated an acceptable path.
        status : string
            A text description of the current status of the propagation.

        """
        status = 'Propagate w/internal engine'
        logger.debug(status)
        success = False
        # Copy the system, so that we can propagate without altering it:
        system = ensemble['system']
        system.potential_and_force()  # Make sure forces are set.
        left, _, right = ensemble['interfaces']
        for i in range(path.maxlen):
            system.order = self.calculate_order(ensemble)
            ekin = calculate_kinetic_energy(system.particles)[0]
            system.particles.ekin = ekin
            status, success, stop, _ = self.add_to_path(path, system.copy(),
                                                        left, right)
            if stop:
                logger.debug('Stopping propagate at step: %i ', i)
                break
            if reverse:
                system.particles.vel *= -1.0
                self.integration_step(system)
                system.particles.vel *= -1.0
            else:
                self.integration_step(system)
        logger.debug('Propagate done: "%s" (success: %s)', status, success)
        return success, status

    @staticmethod
    def modify_velocities(ensemble, vel_settings):
        """Modify the velocities of the current state.

        This method will modify the velocities of a time slice.
        And it is part of the integrator since it, conceptually,
        fits here:  we are acting on the system and modifying it.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `rgen` : object like :py:class:`.RandomGenerator`
              This is the random generator that will be used.

        vel_settings: dict.
            It contains all the info for the velocity:

            * `sigma_v` : numpy.array, optional
              These values can be used to set a standard deviation (one
              for each particle) for the generated velocities.
            * `aimless` : boolean, optional
              Determines if we should do aimless shooting or not.
            * `momentum` : boolean, optional
              If True, we reset the linear momentum to zero after
              generating.
            * `rescale or rescale_energy` : float, optional
              In some NVE simulations, we may wish to re-scale the
              energy to a fixed value. If `rescale` is a float > 0,
              we will re-scale the energy (after modification of
              the velocities) to match the given float.

        Returns
        -------
        dek : float
            The change in the kinetic energy.
        kin_new : float
            The new kinetic energy.

        """
        rgen = ensemble['rgen']
        system = ensemble['system']
        rescale = vel_settings.get('rescale_energy',
                                   vel_settings.get('rescale'))
        particles = system.particles
        if rescale is not None and rescale is not False:
            if rescale > 0:
                kin_old = rescale - particles.vpot
                do_rescale = True
            else:
                logger.warning('Ignored re-scale 6.2%f < 0.0.', rescale)
                return 0.0, calculate_kinetic_energy(particles)[0]
        else:
            kin_old = calculate_kinetic_energy(particles)[0]
            do_rescale = False
        if vel_settings.get('aimless', False):
            vel, _ = rgen.draw_maxwellian_velocities(system)
            particles.vel = vel
        else:  # Soft velocity change, from a Gaussian distribution:
            dvel, _ = rgen.draw_maxwellian_velocities(
                system, sigma_v=vel_settings['sigma_v'])
            particles.vel = particles.vel + dvel
        if vel_settings.get('momentum', False):
            reset_momentum(particles)
        if do_rescale:
            system.rescale_velocities(rescale)
        kin_new = calculate_kinetic_energy(particles)[0]
        dek = kin_new - kin_old
        return dek, kin_new

    def __call__(self, system):
        """To allow calling `MDEngine(system)`.

        Here, we are just calling `self.integration_step(system)`.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we are integrating.

        Returns
        -------
        out : None
            Does not return anything, but will update the particles.

        """
        return self.integration_step(system)

    @staticmethod
    def calculate_order(ensemble, xyz=None, vel=None, box=None):
        """Return the order parameter.

        This method is just to help to calculate the order parameter
        in cases where only the engine can do it.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system` : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.

        xyz : numpy.array, optional
            The positions to use. Typically for internal engines, this
            is not needed. It is included here as it can be used for
            testing and also to be compatible with the generic function
            defined by the parent.
        vel : numpy.array, optional
            The velocities to use.
        box : numpy.array, optional
            The current box vectors.

        Returns
        -------
        out : list of floats
            The calculated order parameter(s).

        """
        system = ensemble['system']
        order_function = ensemble['order_function']
        if any((xyz is None, vel is None, box is None)):
            return order_function.calculate(system)
        system.particles.pos = xyz
        system.particles.vel = vel
        system.box.update_size(box)
        return order_function.calculate(system)

    def kick_across_middle(self, ensemble, middle, tis_settings):
        """Force a phase point across the middle interface.

        This is accomplished by repeatedly kicking the phase point so
        that it crosses the middle interface.

        Parameters
        ----------
        ensemble : dict
            It contains:

            * `system` : object like :py:class:`.System`
              This is the system that contains the particles we are
              investigating
            * `order_function` : object like :py:class:`.OrderParameter`
              The object used for calculating the order parameter.
            * `rgen` : object like :py:class:`.RandomGenerator`
              This is the random generator that will be used.

        middle : float
            This is the value for the middle interface.
        tis_settings : dict
            This dictionary contains settings for TIS. Explicitly used here:

            * `zero_momentum`: boolean, determines if the momentum is zeroed.
            * `rescale_energy`: boolean, determines if energy is re-scaled.

        Returns
        -------
        out[0] : object like :py:class:`.System`
            The phase-point just before the interface.
        out[1] : object like :py:class:`.System`
            The phase-point just after the interface.

        Note
        ----
        This function will update the system state.

        """
        # We search for crossing with the middle interface and do this
        # by sequentially kicking the initial phase point:
        system = ensemble['system']
        previous = system.copy()
        system.potential_and_force()  # Make sure forces are set.
        curr = self.calculate_order(ensemble)[0]
        logger.info('Kicking from: %9.6f', curr)
        while True:
            # Save current state:
            previous = system.copy()
            # Modify velocities:
            self.modify_velocities(
                ensemble, {'sigma_v': None,
                           'aimless': True,
                           'momentum': tis_settings['zero_momentum'],
                           'rescale': tis_settings['rescale_energy']})
            # Update order parameter in case it is velocity dependent:
            curr = self.calculate_order(ensemble)[0]
            previous.order = curr
            # Store modified velocities:
            previous.particles.set_vel(system.particles.get_vel())
            # Integrate forward one step:
            self.integration_step(system)
            # Compare previous order parameter and the new one:
            prev = curr
            curr = self.calculate_order(ensemble)[0]
            if curr == middle:
                # By construction we want two points, one left and one
                # right of the interface, and these two points should
                # be connected by a MD step. If we hit exactly on the
                # interface we just fall back:
                system.particles = previous.particles.copy()
                curr = previous.order
            else:
                if (prev <= middle < curr) or (curr < middle <= prev):
                    # Middle interface was crossed, just stop the loop.
                    logger.info('Crossing found: %9.6f %9.6f ', prev, curr)
                    break
                if (prev <= curr < middle) or (middle < curr <= prev):
                    # We are getting closer, keep the new point.
                    pass
                else:  # We did not get closer, fall back to previous point.
                    system.particles = previous.particles.copy()
                    curr = previous.order
        system.order = curr
        return previous, system

    def dump_phasepoint(self, phasepoint, deffnm=None):
        """For compatibility with external integrators."""
        return

    def clean_up(self):
        """Clean up after using the engine.

        Currently, this is only included for compatibility with external
        integrators.

        """
        return


class Verlet(MDEngine):
    """The Verlet MD integrator.

    This class defines the Verlet MD integrator.

    Attributes
    ----------
    half_idt : float
        Half of the inverse time step: `0.5 / timestep`.
    timestepsq : float
        Squared time step: `timestep**2`.
    previous_pos : numpy.array
        Stores the previous positions of the particles.

    """

    def __init__(self, timestep):
        """Set up the Verlet MD integrator.

        Parameters
        ----------
        timestep : float
            The time step in internal units.

        """
        super().__init__(timestep, 'Verlet MD integrator', dynamics='NVE')
        self.half_idt = 0.5 / self.timestep
        self.timestepsq = self.timestep**2
        self.previous_pos = None

    def set_initial_positions(self, particles):
        """Get initial positions for the Verlet integration.

        Parameters
        ----------
        particles : object like :py:class:`.Particles`
            The initial configuration. Positions and velocities are
            required.

        """
        self.previous_pos = particles.pos - particles.vel * self.timestep

    def integration_step(self, system):
        """Perform one Verlet integration step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            system.

        """
        particles = system.particles
        acc = particles.force * particles.imass
        pos = 2.0 * particles.pos - self.previous_pos + acc * self.timestepsq
        particles.vel = (pos - self.previous_pos) * self.half_idt
        self.previous_pos, particles.pos = particles.pos, pos
        system.potential_and_force()

    def restart_info(self):
        """Return restart info.

        The restart info for the engine Verlet.

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['class'] = 'Verlet'
        for attr, value in self.__dict__.items():
            info[attr] = value

        return info

    def load_restart_info(self, info):
        """Load restart information.

        Parameters
        ----------
        info : dict
            The dictionary with the restart information, should be
            similar to the dict produced by :py:func:`.restart_info`.

        """
        super().load_restart_info(info)
        self.timestep = info.get('timestep')
        self.half_idt = info.get('half_idt')
        self.timestepsq = info.get('timestepsq')
        self.previous_pos = info.get('previous_pos')


class RandomWalk(MDEngine):
    """A Random Walker integrator.

    This class defines a Random walker integrator.

    Attributes
    ----------
    timestep : float
        The length of the step.

    """

    def __init__(self, timestep, rgen=None, seed=0):
        """Set up the Random walker integrator.

        Parameters
        ----------
        timestep : float
            The time step in internal units.
        rgen : string, optional
            This string can be used to pick a particular random
            generator, which is useful for testing.
        seed : integer, optional
            A seed for the random generator.

        """
        super().__init__(timestep, 'Random Walker integrator',
                         dynamics='NoNe')
        rgen_settings = {'seed': seed, 'rgen': rgen}
        self.rgen = create_random_generator(rgen_settings)

    def integration_step(self, system):
        """Random Walker integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        particles.vel, _ = self.rgen.draw_maxwellian_velocities(system)
        particles.pos += self.timestep * particles.vel


class VelocityVerlet(MDEngine):
    """The Velocity Verlet MD integrator.

    This class defines the Velocity Verlet integrator.

    Attributes
    ----------
    half_timestep : float
        Half of the timestep.

    """

    def __init__(self, timestep):
        """Set up the Velocity Verlet integrator.

        Parameters
        ----------
        timestep : float
            The time step in internal units.

        """
        super().__init__(timestep, 'Velocity Verlet MD integrator',
                         dynamics='NVE')
        self.half_timestep = self.timestep * 0.5

    def integration_step(self, system):
        """Velocity Verlet integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        imass = particles.imass
        particles.vel += self.half_timestep * particles.force * imass
        particles.pos += self.timestep * particles.vel
        system.potential_and_force()
        particles.vel += self.half_timestep * particles.force * imass

    def restart_info(self):
        """Return some restart info.

        The restart info for the engine VelocityVerlet.

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['class'] = 'VelocityVerlet'
        for attr, value in self.__dict__.items():
            info[attr] = value

        return info

    def load_restart_info(self, info):
        """Load restart information.

        Parameters
        ----------
        info : dict
            The dictionary with the restart information, should be
            similar to the dict produced by :py:func:`.restart_info`.

        """
        super().load_restart_info(info)
        self.timestep = info.get('timestep')
        self.half_timestep = info.get('half_timestep')
        self.dynamics = info.get('dynamics')


class Langevin(MDEngine):
    """The Langevin MD integrator.

    This class defines a Langevin integrator.

    Attributes
    ----------
    rgen : object like :py:class:`.RandomGenerator`
        This is the class that handles the generation of random numbers.
    gamma : float
        The friction parameter.
    high_friction : boolean
        Determines if we are in the high friction limit and should
        do the over-damped version.
    init_params : boolean
        If true, we will initiate parameters for the Langevin
        integrator when `integrate_step` is invoked.
    param_high : dict
        This contains the parameters for the high friction limit. Here
        we integrate the equations of motion according to:
        ``r(t + dt) = r(t) + dt * f(t)/m*gamma + dr``.
        The items in the dict are:

        * `sigma` : float
          standard deviation for the positions, used when drawing dr.
        * `bddt` : numpy.array
          Equal to ``dt*gamma/masses``, since the masses is a
          numpy.array this will have the same shape.

    param_iner : dict
        This dict contains the parameters for the non-high friction
        limit where we integrate the equations of motion according to:
        ``r(t + dt) = r(t) + c1 * dt * v(t) + c2*dt*dt*a(t) + dr``
        and
        ``v(r + dt) = c0 * v(t) + (c1-c2)*dt*a(t) + c2*dt*a(t+dt) + dv``.
        The dict contains:

        * `c0` : float
          Corresponds to ``c0`` in the equation above.
        * `a1` : float
          Corresponds to ``c1*dt`` in the equation above.
        * `a2` : numpy.array
          Corresponds to ``c2*dt*dt/mass`` in the equation above.
          Here we divide by the masses in order to use the forces rather
          than the acceleration. Since the masses might be different for
          different particles, this will result in a numpy.array with
          shape equal to the shape of the masses.
        * `b1` : numpy.array
          Corresponds to ``(c1-c2)*dt/mass`` in the equation above.
          Here we also divide by the masses, resulting in a numpy.array.
        * `b2` : numpy.array
          Corresponds to ``c2*dt/mass`` in the equation above.
          Here we also divide by the masses, resulting in a numpy.array.
        * `mean` : numpy.array (2,)
          The means for the bivariate Gaussian distribution.
        * `cov` : numpy.array (2,2)
          This array contains the covariance for the bivariate Gaussian
          distribution. `param_iner['mean']` and `param_iner['cov']` are
          used as parameters when drawing ``dr`` and ``dv`` from the
          bivariate distribution.

    Note
    ----
    Currently, we are using a multi-normal distribution from numpy.
    Consider replacing this one as it seems somewhat slow.

    """

    def __init__(self, timestep, gamma, rgen=None, seed=0,
                 high_friction=False):
        """Set up the Langevin integrator.

        Actually, it is very convenient to set some variables for the
        different particles. However, to have a uniform initialisation
        for the different integrators, we postpone this.
        This initialisation can be done later by calling explicitly the
        function `self._init_parameters(system)` or it will be called
        the first time `self.integration_step` is invoked.

        Parameters
        ----------
        timestep : float
            The time step in internal units.
        gamma : float
            The gamma parameter for the Langevin integrator.
        rgen : string, optional
            This string can be used to pick a particular random
            generator, which is useful for testing.
        seed : integer, optional
            A seed for the random generator.
        high_friction : boolean, optional
            Determines if we are in the high_friction limit and should
            do the over-damped version.

        """
        super().__init__(timestep, 'Langevin MD integrator',
                         dynamics='stochastic')
        self.gamma = gamma
        self.high_friction = high_friction
        rgen_settings = {'seed': seed, 'rgen': rgen}
        self.rgen = create_random_generator(rgen_settings)
        self.param_high = {'sigma': None, 'bddt': None}
        self.param_iner = {'c0': None, 'a1': None, 'a2': None,
                           'b1': None, 'b2': None, 'mean': None, 'cov': None}
        self.init_params = True

    def _init_parameters(self, system):
        """Extra initialisation of the Langevin integrator.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but updates ``self.param``.

        """
        beta = system.temperature['beta']
        imasses = system.particles.imass
        if self.high_friction:
            self.param_high['sigma'] = np.sqrt(2.0 * self.timestep *
                                               imasses/(beta * self.gamma))
            self.param_high['bddt'] = self.timestep * imasses / self.gamma
        else:
            gammadt = self.gamma * self.timestep
            exp_gdt = np.exp(-gammadt)
            if self.gamma > 0.0:
                c_0 = exp_gdt
                c_1 = (1.0 - c_0) / gammadt
                c_2 = (1.0 - c_1) / gammadt
            else:
                raise ValueError(
                    f'Langevin: Found gamma = {self.gamma:6.2f} < 0.'
                )

            self.param_iner['c0'] = c_0
            self.param_iner['a1'] = c_1 * self.timestep
            self.param_iner['a2'] = c_2 * self.timestep**2 * imasses
            self.param_iner['b1'] = (c_1 - c_2) * self.timestep * imasses
            self.param_iner['b2'] = c_2 * self.timestep * imasses

            self.param_iner['mean'] = []
            self.param_iner['cov'] = []
            self.param_iner['cho'] = []

            for imass in imasses:
                sig_ri2 = ((self.timestep * imass[0] / (beta * self.gamma)) *
                           (2. - (3. - 4.*exp_gdt + exp_gdt**2) / gammadt))
                sig_vi2 = (1.0 - exp_gdt**2) * imass[0] / beta
                cov_rvi = (imass[0]/(beta * self.gamma)) * (1.0 - exp_gdt)**2
                cov_matrix = np.array([[sig_ri2, cov_rvi],
                                       [cov_rvi, sig_vi2]])
                self.param_iner['cov'].append(cov_matrix)
                self.param_iner['cho'].append(np.linalg.cholesky(cov_matrix))
                self.param_iner['mean'].append(np.zeros(2))

    def integration_step(self, system):
        """Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            system.

        """
        if self.init_params:
            self._init_parameters(system)
            self.init_params = False
        if self.high_friction:
            return self.integration_step_overdamped(system)
        return self.integration_step_inertia(system)

    def integration_step_overdamped(self, system):
        """Over damped Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        system.force()  # update forces
        particles = system.particles
        rands = self.rgen.normal(loc=0.0, scale=self.param_high['sigma'],
                                 size=particles.vel.shape)
        particles.pos += self.param_high['bddt'] * particles.force + rands
        particles.vel = rands
        system.potential()

    def integration_step_inertia(self, system):
        """Langevin integration, one time step.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system to integrate/act on. Assumed to have a particle
            list in ``system.particles``.

        Returns
        -------
        out : None
            Does not return anything, but alters the state of the given
            `system`.

        """
        particles = system.particles
        pos_rand = np.zeros(particles.pos.shape)
        vel_rand = np.zeros(particles.vel.shape)
        if self.gamma > 0.0:
            mean, cov = self.param_iner['mean'], self.param_iner['cov']
            cho = self.param_iner['cho']
            for i, (meani, covi, choi) in enumerate(zip(mean, cov, cho)):
                randxv = self.rgen.multivariate_normal(meani, covi, cho=choi,
                                                       size=system.get_dim())
                pos_rand[i] = randxv[:, 0]
                vel_rand[i] = randxv[:, 1]
        particles.pos += (self.param_iner['a1'] * particles.vel +
                          self.param_iner['a2'] * particles.force + pos_rand)

        vel2 = (self.param_iner['c0'] * particles.vel +
                self.param_iner['b1'] * particles.force + vel_rand)

        system.force()  # Update forces.

        particles.vel = vel2 + self.param_iner['b2'] * particles.force
        system.potential()

    def restart_info(self):
        """Return restart info.

        The restart info for the engine Langevin.

        Returns
        -------
        info : dict
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()
        info['class'] = 'Langevin'
        for attr, value in self.__dict__.items():
            if attr == 'rgen':
                info[attr] = value.get_state()
            else:
                info[attr] = value
        return info

    def load_restart_info(self, info):
        """Load restart information.

        Parameters
        ----------
        info : dict
            The dictionary with the restart information, should be
            similar to the dict produced by :py:func:`.restart_info`.

        """
        super().load_restart_info(info)
        self.timestep = info.get('timestep')
        self.dynamics = info.get('dynamics')
        self.gamma = info.get('gamma')
        self.high_friction = info.get('high_friction')
        self.param_high = info.get('param_high')
        self.param_iner = info.get('param_iner')
        self.init_params = info.get('init_params')

        if 'rgen' in info:
            self.rgen = create_random_generator(info['rgen'])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""The sub-package handles input and output for PyRETIS.

This package is intended for creating various forms of output
from the PyRETIS program. It includes writers for simple text-based
output and plotters for creating figures. Figures and the text results
can be combined into reports, which are handled by the report module.

Package structure
~~~~~~~~~~~~~~~~~

Modules
~~~~~~~

__init__.py
    Imports from the other modules.

checker.py (:py:mod:`pyretis.inout.checker`)
    Functions to check the compatibility and
    consistency of various input values.

common.py (:py:mod:`pyretis.inout.common`)
    Common functions and variables for the input/output. These
    functions are mainly intended for internal use and are not imported
    here.

fileio.py (:py:mod:`pyretis.inout.fileio`)
    A module which defines a generic file class for PyRETIS output files.

settings.py (:py:mod:`pyretis.inout.settings`)
    A module which handles the reading/writing of settings.

simulationio.py (:py:mod:`pyretis.inout.simulationio`)
    A module which handles th reading/writing of the main simulations.

restart.py (:py:mod:`pyretis.inout.restart`)
    A module which handles restart reading/writing.

Sub-packages
~~~~~~~~~~~~

analysisio (:py:mod:`pyretis.inout.analysisio`)
    Handles the input and output needed for analysis.

formats (:py:mod:`pyretis.inout.formats`)
    Handles the input and output of different data formats. This
    includes the configurations and the internal data formats.

plotting (:py:mod:`pyretis.inout.plotting`)
    Handles the plotting needed by the analysis by defining plotting
    tools, methods and styles.

report (:py:mod:`pyretis.inout.report`)
    Generate reports with results from simulations.


Important classes defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Important methods defined in this package
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_report (:py:func:`.generate_report`)
    A function to generate reports from analysis output(s).

parse_settings_file (:py:func:`.parse_settings_file`)
    Method for parsing settings from a given input file.

write_settings_file (:py:func:`.write_settings_file`)
    Method for writing settings from a simulation to a given file.

write_restart_file (:py:func:`.write_restart_file`)
    Method for writing restart information.
"""
from pyretis.inout.screen import print_to_screen



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module defines how we write and read restart files.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

read_restart_file (:py:func:`.read_restart_file`)
    A method for reading restart information from a file.

write_restart_file (:py:func:`.write_restart_file`)
    A method for writing the restart file.

write_ensemble_restart (:py:func:`.write_ensemble_restart`)
    A method for writing restart files for path ensembles.
"""
import logging
import copy
import os
import pickle
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['read_restart_file',
           'write_restart_file',
           'write_ensemble_restart']


def write_restart_file(filename, simulation):
    """Write restart info for a simulation.

    Parameters
    ----------
    filename : string
        The file we are going to write to.
    simulation : object like :py:class:`.Simulation`
        A simulation object we will get information from.

    """
    info = simulation.restart_info()

    with open(filename, 'wb') as outfile:
        pickle.dump(info, outfile)


def write_ensemble_restart(ensemble, settings_ens):
    """Write a restart file for a path ensemble.

    Parameters
    ----------
    ensemble : dict
        it contains:

        * `path_ensemble` : object like :py:class:`.PathEnsemble`
          The path ensemble we are writing restart info for.
        * ` system` : object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function` : object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine` : object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    settings_ens : dict
        A dictionary with the ensemble settings.

    """
    info = copy.deepcopy(settings_ens)
    if ensemble.get('path_ensemble') is not None:
        info['path_ensemble'] = ensemble['path_ensemble'].restart_info()
    if ensemble.get('system') is not None:
        info['system'] = ensemble['system'].restart_info()
    if ensemble.get('engine') is not None:
        info['engine'] = ensemble['engine'].restart_info()
    if ensemble.get('rgen') is not None:
        info['rgen'] = ensemble['rgen'].get_state()
    if ensemble.get('order_function') is not None:
        info['order_function'] = ensemble['order_function'].restart_info()

    filename = os.path.join(
        settings_ens['simulation']['exe_path'],
        ensemble['path_ensemble'].ensemble_name_simple,
        'ensemble.restart')
    with open(filename, 'wb') as outfile:
        pickle.dump(info, outfile)


def read_restart_file(filename):
    """Read restart info for a simulation.

    Parameters
    ----------
    filename : string
        The file we are going to read from.

    """
    with open(filename, 'rb') as infile:
        info = pickle.load(infile)
    return info



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module handles the set-up of initial positions and a box.

The initial positions can either be generated on a lattice, or it can
be read from a file.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

set_up_box (:py:func:`.set_up_box`)
    Create a simulation box from simulation settings.

create_initial_positions (:py:func:`.create_initial_positions`)
    Get initial positions based on settings. This will either be
    read from a file or generated on a lattice.

create_system (:py:func:`.create_system`)
    Set up a system from given settings. This method will probably
    also need to set/get the initial positions and velocities for
    the particles and set up the simulation box.

create_velocities (:py:func:`.create_velocities`)
    Create velocities from settings for a system with particles.

initial_positions_file (:py:func:`.initial_positions_file`)
    Get initial positions from a file.

initial_positions_lattice (:py:func:`.initial_positions_lattice`)
    Get initial positions by generating a lattice.

"""
import logging
import os
import numpy as np
from pyretis.tools import generate_lattice
from pyretis.core.box import create_box
from pyretis.core.system import System
from pyretis.core.particles import Particles, get_particle_type
from pyretis.core.units import CONVERT
from pyretis.inout.settings import look_for_input_files
from pyretis.inout.formats.snapshot import read_txt_snapshots
from pyretis.inout.formats.xyz import read_xyz_file
from pyretis.inout.formats.gromacs import read_gromacs_file, read_gromos96_file
from pyretis.engines.openmm import make_pyretis_system as openmm_system
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['create_initial_positions', 'create_system', 'create_velocities',
           'set_up_box', 'initial_positions_file', 'initial_positions_lattice']


PERIODIC_TABLE = {'H': 1.007975, 'He': 4.002602, 'Li': 6.9675,
                  'Be': 9.0121831, 'B': 10.8135, 'C': 12.0106,
                  'N': 14.006855, 'O': 15.9994, 'F': 18.998403163,
                  'Ne': 20.1797, 'Na': 22.98976928, 'Mg': 24.3055,
                  'Al': 26.9815385, 'Si': 28.085, 'P': 30.973761998,
                  'S': 32.0675, 'Cl': 35.4515, 'Ar': 39.948,
                  'K': 39.0983, 'Ca': 40.078, 'Sc': 44.955908,
                  'Ti': 47.867, 'V': 50.9415, 'Cr': 51.9961,
                  'Mn': 54.938044, 'Fe': 55.845, 'Co': 58.933194,
                  'Ni': 58.6934, 'Cu': 63.546, 'Zn': 65.38,
                  'Ga': 69.723, 'Ge': 72.63, 'As': 74.921595,
                  'Se': 78.971, 'Br': 79.904, 'Kr': 83.798,
                  'Rb': 85.4678, 'Sr': 87.62, 'Y': 88.90584,
                  'Zr': 91.224, 'Nb': 92.90637, 'Mo': 95.95,
                  'Ru': 101.07, 'Rh': 102.9055, 'Pd': 106.42,
                  'Ag': 107.8682, 'Cd': 112.414, 'In': 114.818,
                  'Sn': 118.71, 'Sb': 121.76, 'Te': 127.6,
                  'I': 126.90447, 'Xe': 131.293, 'Cs': 132.90545196,
                  'Ba': 137.327, 'La': 138.90547, 'Ce': 140.116,
                  'Pr': 140.90766, 'Nd': 144.242, 'Sm': 150.36,
                  'Eu': 151.964, 'Gd': 157.25, 'Tb': 158.92535,
                  'Dy': 162.5, 'Ho': 164.93033, 'Er': 167.259,
                  'Tm': 168.93422, 'Yb': 173.045, 'Lu': 174.9668,
                  'Hf': 178.49, 'Ta': 180.94788, 'W': 183.84,
                  'Re': 186.207, 'Os': 190.23, 'Ir': 192.217,
                  'Pt': 195.084, 'Au': 196.966569, 'Hg': 200.592,
                  'Tl': 204.3835, 'Pb': 207.2, 'Bi': 208.9804,
                  'Th': 232.0377, 'Pa': 231.03588, 'U': 238.02891}


# Variable that defines some information for reading input files:
# TODO: Move this to the place where the file readers are defined?
READFILE = {'xyz': {'reader': read_xyz_file,
                    'units': {'length': 'A', 'velocity': 'A/fs'}},
            'txt': {'reader': read_txt_snapshots,
                    'units': None},
            'gro': {'reader': read_gromacs_file,
                    'units': {'length': 'nm', 'velocity': 'nm/ps'}},
            'g96': {'reader': read_gromos96_file,
                    'units': {'length': 'nm', 'velocity': 'nm/ps'}}}


def list_get(input_list, index):
    """Get an item from a list and handle out-of bounds errors.

    This method is intended to be used when we are picking items from
    a list and possibly we want a number of items which is larger than
    the number of items in the list. Here, we then just return the last
    element.

    Parameters
    ----------
    input_list : list
        The list to pick from.
    index : integer
        The index to pick.

    """
    try:
        return input_list[index]
    except IndexError:
        return input_list[-1]


def _assign_mass_from_file(filename, unit):
    """Assign masses and imasses to particles from a configuration file.

    This function is useful when velocities will be internally modified
    while using an external engine.

    Parameters
    ----------
    filename : string
        The configuration file.
    unit : string
        The system of units.

    """
    snapshot, _ = _get_snapshot_from_file({'input_file': filename}, unit)
    masses = []
    for i, atomname in enumerate(snapshot['atomname']):
        masses.append([guess_particle_mass(i+1, atomname, unit)*1.])
    return masses, np.reciprocal(masses)


def guess_particle_mass(particle_no, particle_type, unit):
    """Guess a particle mass from it's type.

    Parameters
    ----------
    particle_no : integer
        Just used to identify the particle number.
    particle_type : string
        Used to identify the particle.
    unit : string
        The system of units. This is used in case we try to get the
        mass from the periodic table where the units are in `g/mol`.

    """
    logger.info(('Mass not specified for particle no. %i\n'
                 'Will guess from particle type "%s"'), particle_no,
                particle_type)
    mass = PERIODIC_TABLE.get(particle_type, None)
    if mass is None:
        particle_mass = 1.0
        logger.info(('-> Could not find mass. '
                     'Assuming %f (internal units)'), particle_mass)
    else:
        particle_mass = CONVERT['mass']['g/mol', unit] * mass
        logger.info(('-> Using a mass of %f g/mol '
                     '(%f in internal units)'), mass, particle_mass)
    return particle_mass


def initial_positions_lattice(settings):
    """Generate initial positions based on given settings.

    We assume here the input values are given with the correct units
    as dictated by ``settings['system']['units']``.

    Parameters
    ----------
    settings : dict
        The input settings for the simulation.

    Returns
    -------
    particles : object like :py:class:`.Particles`
        The particles we created.
    size : list of floats
        A size for the region we created. This can be used to create
        a box.

    """
    pos_settings = settings['particles']['position']
    ptype = settings['particles'].get('ptype', [0])
    pname = settings['particles'].get('name', ['Ar'])

    lattice, size = generate_lattice(
        pos_settings['generate'].lower(),
        pos_settings.get('repeat'),
        lcon=pos_settings.get('lcon', None),
        density=pos_settings.get('density', None)
    )

    ndim = settings['system'].get('dimensions', None)
    if ndim is None:
        ndim = len(size)
    else:
        if ndim != len(size):
            msgtxt = ('Inconsistent dimenaions in settings and generated '
                      f'lattice.\nSettings gives {ndim}D while the generated '
                      f'lattice is {len(size)}D.')
            logger.error(msgtxt)
            raise ValueError(msgtxt)

    particles = Particles(dim=ndim)

    for i, pos in enumerate(lattice):
        particle_type = list_get(ptype, i)
        particle_name = list_get(pname, i)
        # Infer the mass from the input masses, or try to get it
        # from the periodic table:
        p_mass = settings['particles'].get('mass', {}).get(particle_name)
        if p_mass is None:
            p_mass = guess_particle_mass(i + 1, particle_name,
                                         settings['system']['units'])
        particles.add_particle(
            pos,
            np.zeros_like(pos),
            np.zeros_like(pos),
            mass=p_mass,
            name=particle_name,
            ptype=particle_type
        )
    logger.info('Generated %i particles on lattice "%s".',
                particles.npart, pos_settings['generate'].lower())
    logger.info('Lattice is %iD.', ndim)

    size = np.array(size)
    box = {'low': size[:, 0], 'high': size[:, 1]}
    return particles, box


def _get_snapshot_from_file(pos_settings, units):
    """Get a configuration snapshot from a file.

    This snapshot will be used to set up the initial configuration.

    Parameters
    ----------
    pos_settings : dict
        A dict with information on what we should read.
    units : string
        The internal units.

    Returns
    -------
    snapshot : dict
        The snapshot we found in the file. It will at least have the
        keys with the positions ('x', 'y', 'z') and atom name
        'atomnames'. It may have information about velocities
        ('vx', 'vy', 'vz') and the box ('box').
    convert : dict
        Dictionary with conversion factors to internal units.

    """
    filename = pos_settings.get('input_file', None)
    if filename is None:
        msg = ('Requested reading (initial) configuration from file, '
               'but no input file given!')
        logger.error(msg)
        raise ValueError(msg)
    fmt = pos_settings.get('format', os.path.splitext(filename)[1][1:])
    snaps = []
    convert = None
    if fmt not in READFILE:
        msg = (f'Input configuration "{filename}" has unknown '
               f'format "{fmt}".')
        logger.error(msg)
        logger.error('Supported formats are: %s.', list(READFILE))
        raise ValueError(msg)

    reader = READFILE[fmt]['reader']
    read_units = READFILE[fmt]['units']
    if read_units is None:
        convert = {'length': 1.0, 'velocity': 1.0}
    else:
        convert = {
            'length': CONVERT['length'][read_units['length'], units],
            'velocity': CONVERT['velocity'][read_units['velocity'], units]
        }
    logger.info(
        'Reading initial configuration from "%s" (format: "%s").',
        filename,
        fmt,
    )
    snaps = [snap for snap in reader(filename)]

    snapshot = None
    if len(snaps) == 1:
        snapshot = snaps[0]
    elif len(snaps) > 1:
        msg = ('Found several frames ({}) in input file.'
               ' Will use the last one!').format(len(snaps))
        logger.warning(msg)
        snapshot = snaps[-1]
    else:
        msg = f"Could not find any configurations in input file: {filename}"
        logger.error(msg)
        raise ValueError(msg)
    return snapshot, convert


def initial_positions_file(settings):
    """Get initial positions from an input file.

    Parameters
    ----------
    settings : dict
        The input settings for the simulation.

    Returns
    -------
    particles : object like :py:class:`.Particles`
        The particles we created.
    size : list of floats
        A size for the region we created. This can be used to create
        a box.
    vel_read : boolean
        True if we read velocities from the input file.

    """
    ndim = settings['system'].get('dimensions', 3)
    pos_settings = settings['particles']['position']
    ptype = settings['particles'].get('ptype', None)
    pname = settings['particles'].get('name', None)
    pmass = settings['particles'].get('mass', {})
    ptypes = {}  # To automatically set particle types based on name.
    snapshot, convert = _get_snapshot_from_file(pos_settings,
                                                settings['system']['units'])
    vel_read = False
    particles = Particles(dim=ndim)
    for i, atomname in enumerate(snapshot['atomname']):
        pos = []
        vel = []
        for key in ['x', 'y', 'z'][:ndim]:
            pos.append(snapshot[key][i])
            vel_key = f'v{key}'
            if vel_key in snapshot:
                vel.append(snapshot[vel_key][i])
        pos = np.array(pos) * convert['length']
        if len(vel) != ndim:
            vel = np.zeros_like(pos)
        else:
            vel = np.array(vel) * convert['velocity']
            vel_read = True
        # Get particle type from the atom names or from input list:
        if ptype is None:
            if atomname not in ptypes:
                ptypes[atomname] = len(ptypes)
            particle_type = ptypes[atomname]
        else:
            particle_type = list_get(ptype, i)
        if pname is None:
            particle_name = atomname
        else:
            particle_name = list_get(pname, i)
        # Infer the mass from the input masses, or try to get it
        # from the periodic table:
        try:
            particle_mass = pmass[particle_name]
        except KeyError:
            particle_mass = guess_particle_mass(i + 1, particle_name,
                                                settings['system']['units'])
        particles.add_particle(pos,
                               vel, np.zeros_like(pos),
                               mass=particle_mass, name=particle_name,
                               ptype=particle_type)
    try:
        box = {'cell': [i * convert['length'] for i in snapshot['box']]}
        if ndim < 3:
            box['cell'] = box['cell'][:ndim]
    except (KeyError, IndexError, TypeError) as err:
        logger.debug('No box read from file: %s.', err)
        box = None
    logger.info('Read %d particle(s) from "%s".', particles.npart,
                pos_settings['input_file'])
    if vel_read:
        logger.info('Read velocities from file: "%s".',
                    pos_settings['input_file'])
    return particles, box, vel_read


def create_initial_positions(settings):
    """Set up the initial positions from the given settings.

    The settings can specify the initial positions as a file or
    to be generated on a lattice by PyRETIS.

    Parameters
    ----------
    settings : dict
        Settings for creating the initial positions.

    Returns
    -------
    out[0] : object like :py:class:`.Particles`
        The particles we created.
    out[1] : list
        The size associated with the particles. Can be used to create a
        box.
    out[2] : boolean
        True if we have read/created velocities different from just
        zeros. This is only True if we have read from a file with
        velocities.

    """
    logger.debug('Settings used for initial positions: %s',
                 settings['particles']['position'])
    particles = None
    if 'generate' in settings['particles']['position']:
        particles, box = initial_positions_lattice(settings)
        return particles, box, False
    if 'input_file' in settings['particles']['position']:
        # First check if we need to add a path to the file:
        filename = settings['particles']['position']['input_file']
        if (not os.path.isfile(filename) and
                'exe_path' in settings['simulation']):
            filename = os.path.join(settings['simulation']['exe_path'],
                                    filename)
            settings['particles']['position']['input_file'] = filename
        particles, box, vel = initial_positions_file(settings)
        return particles, box, vel
    msg = 'Unknown settings for initial positions: {}'
    msgtxt = msg.format(settings['particles']['position'])
    logger.error(msgtxt)
    raise ValueError(msgtxt)


def set_up_box(settings, boxs, dim=3):
    """Set up a box from given settings.

    Parameters
    ----------
    settings : dict
        The dict with the simulation settings.
    boxs : dict or None
        If no box settings are given, we can still create a box,
        inferred from the positions of the particles. This dict
        contains the settings to do so.
    dim : integer, optional
        Number of dimensions for the box. This is used only as a last
        resort when no information about the box is given.

    Returns
    -------
    box : object like :py:class:`.BoxBase` or None
        The box if we managed to create it, otherwise None.

    """
    msg = 'Box created {}:\n{}'
    box = None
    if settings.get('box', None) is not None:
        box = create_box(**settings['box'])
        msgtxt = msg.format('from settings', box)
        logger.info(msgtxt)
        debugtxt = f"Settings used:\n{settings['box']}"
        logger.debug(debugtxt)
    else:
        if boxs is not None:
            box = create_box(**boxs)
            msgtxt = msg.format('from initial positions', box)
            logger.info(msgtxt)
            msgwarn = 'The box was assumed periodic in all directions.'
            logger.warning(msgwarn)
        else:
            if dim > 0:
                box = create_box(periodic=[False]*dim)
                msgtxt = msg.format('without specifications', box)
                logger.info(msgtxt)
                msgwarn = 'The box was assumed nonperiodic in all directions.'
                logger.warning(msgwarn)
    return box


def create_velocities(system, settings, vel):
    """Create velocities from settings for a system.

    Parameters
    ----------
    system : object like :py:class:`.System`
        The system to create velocities for. It's needed since
        we need to know the degrees of freedom.
    settings : dict
        Settings to use for creating the velocities.
    vel : boolean
        If True, we already read velocities. They will now be
        overwritten. We just make some warnings about this.

    Returns
    -------
    out : boolean
        True if we actually generated velocities.

    """
    vel_settings = settings['particles'].get('velocity', {})
    if vel:
        logger.info(
            'Velocities read from input configuration (temperature: %6.2g).',
            system.calculate_temperature(),
        )
    if 'generate' in vel_settings:
        if vel:
            logger.warning(
                'Will generate and overwrite velocities already set.'
            )
        gen_settings = {'distribution': vel_settings['generate']}
        for key in ('seed', 'momentum', 'temperature', 'rgen'):
            try:
                gen_settings[key] = vel_settings[key]
            except KeyError:
                pass
        system.generate_velocities(**gen_settings)
        logger.info(
            'Generated new velocities with average temperature: %6.2g',
            system.calculate_temperature(),
        )
        logger.debug('Settings used for generating velocities: %s',
                     gen_settings)
        return True
    if 'scale' in vel_settings:
        target = vel_settings['scale']
        # Just set the velocities to some temperature for now.
        # The scaling is done later by calling `system.extra_setup()`.
        gen_settings = {
            'distribution': 'maxwell',
            'momentum': system.particles.npart != 1,
            'temperature': settings['system'].get('temperature', 1.0),
        }
        system.generate_velocities(**gen_settings)
        msg = f'Scaling velocities to total energy {target}'
        logger.debug(msg)
        system.post_setup.append(('rescale_velocities', (target,)))
        return True
    return False


def create_system(settings):
    """Create a system from input settings.

    Parameters
    ----------
    settings : dict
        The dict with the simulation settings.

    Returns
    -------
    system : object like :py:class:`.System`
        The system object we create here.

    """
    if 'restart' in settings:
        system = System()
        system.load_restart_info(settings['restart']['system'])
        return system

    vel = None
    if ('engine' not in settings or
            settings['engine'].get('type', 'internal') == 'internal'):
        particles, box, vel = create_initial_positions(settings)
        box = set_up_box(settings, box, dim=particles.dim)
    elif settings['engine'].get('type', None).lower() == 'openmm':
        return openmm_system(settings)
    else:
        # Engine is not None and not internal => external.
        klass = get_particle_type('external')
        particles = klass(dim=3)
        if 'input_path' in settings['engine']:
            required_file = {}
            # Get the engine input files
            if 'cp2k' in settings['engine']['class'].lower():
                required_file = {'conf': 'initial.{}'.format(
                    settings['engine'].get('cp2k_format', 'xyz'))}
            elif 'gromacs' in settings['engine']['class'].lower():
                required_file = {
                    'conf':
                        f"conf.{settings['engine'].get('gmx_format', 'gro')}"}

            input_file = look_for_input_files(settings['engine']['input_path'],
                                              required_file)
            particles.set_pos((input_file['conf'], None))
            # gromacs tests fail without this if-statement.
            # _get_snapshot_from_file() has issues with reading .g96 format.
            if settings['engine']['class'].lower() == 'cp2k':
                particles.mass, particles.imass = _assign_mass_from_file(
                    input_file['conf'],
                    settings['system']['units'])
        else:
            particles.set_pos((None, None))
        particles.set_vel(False)
        box = None

    system = System(
        temperature=settings['system']['temperature'],
        units=settings['system']['units'],
        box=box
        )
    system.particles = particles

    # figure out what to do with velocities:
    if 'position' in settings['particles'] and \
            'input_file' not in settings['particles']['position']:
        vel_gen = create_velocities(system, settings, vel)
        if not (vel_gen or vel):
            logger.warning('Velocities not created/read: Set to zero!')

    return system



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Plot results from time tests of the Lennard-Jones potential.

Here we assume that the results are available in files named
``timings.txt``.
"""
import numpy as np
from matplotlib import pylab as plt

RESULTS = {'c': 'c/timings.txt',
           'fortran': 'fortran/timings.txt',
           'fortran (pointer)': 'fortran-pointer/timings.txt',
           'numpy': 'timings-numpy.txt',
           'pure python': 'timings-python.txt'}

MARKERS = ('o', 'v', '^', 's', '*', 'p', 'h', 'x', '+')


if __name__ == '__main__':
    # pylint: disable=invalid-name
    data = {}
    for key in RESULTS:
        data[key] = np.loadtxt(RESULTS[key])

    fig = plt.figure()
    ax1 = fig.add_subplot(111)
    for i, key in enumerate(data):
        result = data[key]
        npart, avg, std = result[:, 0], result[:, 2], result[:, 3]
        ax1.errorbar(npart, avg, yerr=std, label=key, lw=3, markersize=9,
                     marker=MARKERS[i], alpha=0.9)
    ax1.legend(ncol=2, loc='center right')
    ax1.set_xlabel('System size')
    ax1.set_ylabel('Time')
    fig.tight_layout()

    fig2 = plt.figure()
    ax2 = fig2.add_subplot(111)
    norm_avg = data['c'][:, 2]

    for i, key in enumerate(data):
        result = data[key]
        npart, avg = result[:, 0], result[:, 2]
        nmin = min(len(norm_avg), len(avg))
        avg_n = avg[:nmin] / norm_avg[:nmin]
        ax2.plot(npart, avg_n, lw=3, label=key, markersize=9, alpha=0.9,
                 marker=MARKERS[i])
    ax2.legend(ncol=2, loc='best', bbox_to_anchor=(0.3, 0.7))
    ax2.set_yscale('log')
    ax2.set_xlabel('System size')
    ax2.set_ylabel('Time relative to python+c')
    fig2.tight_layout()
    plt.show()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the FileIO class."""
import filecmp
import logging
import os
import shutil
import pathlib
import tempfile
import unittest
from io import StringIO
from unittest.mock import patch
import numpy as np
from numpy.random import randint, rand
from pyretis.core.path import Path
from pyretis.core.pathensemble import PathEnsemble
from pyretis.inout.analysisio.analysisio import run_analysis, read_first_block
from pyretis.inout.formats.formatter import OutputFormatter
from pyretis.inout.formats.snapshot import read_txt_snapshots
from pyretis.inout.fileio import FileIO
from pyretis.inout.formats.energy import EnergyFile, EnergyPathFile
from pyretis.inout.formats.cross import CrossFile
from pyretis.inout.formats.order import OrderFile, OrderPathFile
from pyretis.inout.formats.pathensemble import PathEnsembleFile
from pyretis.inout.formats.path import PathExtFile, PathIntFile
from pyretis.inout.formats.snapshot import SnapshotFile
from pyretis.inout.settings import parse_settings_file
from .help import (
    assert_equal_path_dict,
    create_external_path,
    create_test_paths,
    turn_on_logging,
    remove_file,
    set_up_system,
)

logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))


PATH_DATA = [
    {
        'cycle': 0, 'generated': ['ki', 0.0, 0, 0],
        'interface': ('L', 'M', 'L'), 'length': 494,
        'ordermax': (-0.765171027, 255), 'ordermin': (-0.9002987171, 0),
        'status': 'ACC', 'weight': 1.,
    },
    {
        'cycle': 1, 'generated': ['s-', 0.0, 0, 0],
        'interface': ('L', 'M', 'L'), 'length': 400,
        'ordermax': (-0.8643118362, 136),
        'ordermin': (-0.9004062103, 399),
        'status': 'ACC', 'weight': 1.,
    },
    {
        'cycle': 2, 'generated': ['tr', 0.0, 0, 0],
        'interface': ('L', 'M', 'L'), 'length': 400,
        'ordermax': (-0.8643118362, 263), 'ordermin': (-0.9004062103, 0),
        'status': 'ACC', 'weight': 1.,
    },
]


class TestAnaIo(unittest.TestCase):
    """Test the analysisio files."""
    def test_read_first_block(self):
        """Test read first block function."""
        with tempfile.TemporaryDirectory() as tempdir:
            filegone = os.path.join(tempdir, 'Mordvichev')
            pathlib.Path(filegone).touch(exist_ok=True)
            out = read_first_block(file_type='energy', file_name=filegone)
            self.assertEqual(out, None)
            with self.assertRaises(ValueError) as ext:
                read_first_block(file_type='BT2', file_name=filegone)
            self.assertIn('Unknown file type', str(ext.exception))
            filetest = os.path.join(HERE, 'order-data.txt')
            with turn_on_logging():
                with self.assertLogs(
                        'pyretis.inout.analysisio.analysisio',
                        level='WARNING'):
                    read_first_block(file_type='order', file_name=filetest)

    def test_run_analysis(self):
        """Test analysisio function"""
        lines = ['']*8
        moves = ['ki', 'sh', 'ki', 'sh', 'ld', 'xx', 'sh']
        line = "         2          3          0 L M R      99 ACC "
        line += "sh -9.108621358e-00  1.007510172e+01       0"
        line += "      98  0.000000000e+00       0       0  1.00e+00\n"
        hline = "# Cycle: 1, status: ACC, move: ('sh', -1.430, 18, 14)\n"
        olines = ['#     Time       Orderp\n',
                  '          0     0.11319\n', '          1    -0.1367\n',
                  '          2    -0.78098\n', '          3    -1.0238\n',
                  '          4    -1.11127\n', '          5    -1.2367\n',
                  '          6    -0.58084\n', '          7     0.0238\n',
                  '          8    -1.06811\n', '          9     0.1131\n',
                  '         10     1.06811\n', '         11     0.3131\n']

        elines = ['#     Time      Potential        Kinetic\n',
                  '          0   -1706.522027    1267.214065\n',
                  '          1   -8706.550067    1167.214065\n',
                  '          2  -18720.057124    1780.765874\n',
                  '          3  -18706.550067    1767.214065\n',
                  '          4  -18705.164689    1765.349001\n',
                  '          5  -18629.768664    1691.047800\n',
                  '          6  -18677.061888    1737.016339\n',
                  '          7  -18567.169855    1628.072515\n',
                  '          8  -18515.405500    1576.664809\n',
                  '          9  -18515.405500    1576.664809\n',
                  '         10  -18583.703121    1643.763063\n',
                  '         11  -18583.703121    1643.763063\n']
        elines2 = [
            '# Time       Potential      Kinetic       Total    Temperature\n',
            '   0      -0.165063      0.000000      -0.165063      0.000000\n',
            '   1      -0.164748      0.000853      -0.163895      0.001706\n',
            '   2      -0.164173      0.001763      -0.162411      0.003526\n',
            '   3      -0.163650      0.002079      -0.161571      0.004157\n',
            '   4      -0.163075      0.001212      -0.161863      0.002424\n',
            '   5      -0.162872      0.000026      -0.162847      0.000051\n',
            '   6      -0.163122      0.000634      -0.162489      0.001267\n',
            '   7      -0.163780      0.003669      -0.160110      0.007339\n',
            '   8      -0.164685      0.004810      -0.159874      0.009620\n',
            '   9      -0.164685      0.004810      -0.159874      0.009620\n',
            '  10      -0.164685      0.004810      -0.159874      0.009620\n',
            '  11      -0.174685      0.008810      -0.129874      0.008620\n']

        clines = [' #     Step  Int Dir \n', '         3    1   - \n',
                  '         1    2   +  \n', '         4    1   - \n',
                  '         6    2   +  \n', '         7    1   - \n',
                  '         6    2   +  \n', '         7    1   - \n',
                  '         6    2   -  \n', '         7    1   + \n',
                  '         8    1   -  \n', '         9    1   + \n']

        for j in range(7):
            lines[j] = line.replace('sh', moves[j])
        lines[7] = line.replace('ACC', 'BWI')
        lines[0] = line.replace('L', 'R')
        lines[1] = line.replace('L', 'R')

        input_file = os.path.join(HERE, 'settings-retis.rst')
        settings = parse_settings_file(input_file)
        settings['simulation']['interfaces'] = [-0.9, -0.8, 0]
        with tempfile.TemporaryDirectory() as tempdir:
            for i in ['000', '001', '002']:
                i_folder = os.path.join(tempdir, i)
                os.mkdir(i_folder)
                with open(i_folder + '/pathensemble.txt', 'w',
                          encoding='utf-8') as f_p, \
                        open(i_folder + '/energy.txt', 'w',
                             encoding='utf-8') as f_e, \
                        open(i_folder + '/order.txt', 'w',
                             encoding='utf-8') as f_o:
                    for line in lines:
                        f_p.write(line)
                        f_e.write(hline)
                        for line in elines:
                            f_e.write(line)
                        f_o.write(hline)
                        for line in olines:
                            f_o.write(line)
            settings['simulation']['exe-path'] = tempdir
            settings['analysis']['report-dir'] = tempdir
            with patch('sys.stdout', new=StringIO()) as stdout:
                run_analysis(settings)
            self.assertIn('Pathensemble data', stdout.getvalue().strip())
            self.assertIn('99.000000', stdout.getvalue().strip())

        with tempfile.TemporaryDirectory() as tempdir:
            with open(os.path.join(tempdir, 'pathensemble.txt'), 'w',
                      encoding='utf-8') as f_p, \
                    open(os.path.join(tempdir, 'energy.txt'), 'w',
                         encoding='utf-8') as f_e, \
                    open(os.path.join(tempdir, 'order.txt'), 'w',
                         encoding='utf-8') as f_o, \
                    open(os.path.join(tempdir, 'cross.txt'), 'w',
                         encoding='utf-8') as c_r:
                for line in lines:
                    f_p.write(line)
                for line in elines2:
                    f_e.write(line)
                f_o.write(hline)
                for line in olines:
                    f_o.write(line)
                for line in clines:
                    c_r.write(line)
                inputfile = os.path.join(HERE, 'settings.rst')
            settings = parse_settings_file(inputfile)
            settings['simulation']['exe-path'] = tempdir
            settings['analysis']['report-dir'] = tempdir
            settings['simulation']['task'] = 'md-flux'
            settings['simulation']['endcycle'] = 8
            settings['simulation']['interfaces'] = [-0.9, -0.8]
            with patch('sys.stdout', new=StringIO()) as stdout:
                run_analysis(settings)
            self.assertIn('0.2500', stdout.getvalue().strip())
            self.assertIn('0.3750', stdout.getvalue().strip())

        settings['simulation']['task'] = 'mistery'
        with self.assertRaises(ValueError) as err:
            run_analysis(settings)
            self.assertIn('Unknown ', err.exception)

        # repptis analysis
        with tempfile.TemporaryDirectory() as tempdir:
            # copy pptis/pathensemble.txt files to tempdir
            for i, ens in enumerate(['000', '001', '002', '003']):
                os.mkdir(os.path.join(tempdir, ens))
                shutil.copy(
                    os.path.join(HERE, 'pptis/pathensemble'+str(i)+'.txt'),
                    os.path.join(tempdir, ens, 'pathensemble.txt')
                )
            inputfile = os.path.join(HERE, 'pptis/retis.rst')
            settings = parse_settings_file(inputfile)
            settings['simulation']['exe-path'] = tempdir
            settings['analysis']['report-dir'] = tempdir
            settings['simulation']['task'] = 'repptis'
            with patch('sys.stdout', new=StringIO()) as stdout:
                run_analysis(settings)
            # flux
            self.assertIn('0.862851052', stdout.getvalue().strip())
            # pcross
            self.assertIn('2.213477168e-03', stdout.getvalue().strip())


class TestFileIO(unittest.TestCase):
    """Test the FileIO class."""

    def test_initiation(self):
        """Test initiation and opening of files."""
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.fileio', level='INFO'):
                FileIO('some-name', 'r', None, backup='not-an-option!')

        # Test for completely new file:
        filename = os.path.join(HERE, 'a_new_file')
        remove_file(filename)
        fileio = FileIO(filename, 'w', None, backup=False)
        self.assertIsNone(fileio.fileh)
        fileio.open()
        self.assertIsNotNone(fileio.fileh)
        fileio.write('test')
        fileio.close()
        remove_file(filename)

        # Test when a file exists:
        filename = os.path.join(HERE, 'already_exists2')
        remove_file(filename)
        remove_file(f'{filename}_000')
        with open(filename, 'w', encoding='utf-8') as fileh:
            fileh.write('test')
        fileio = FileIO(filename, 'w', None, backup=True)
        fileio.open()
        fileio.write('text')
        self.assertIsNotNone(fileio.fileh)
        self.assertTrue(os.path.isfile(filename))
        self.assertTrue(os.path.isfile(f'{filename}_000'))
        del fileio
        remove_file(filename)
        remove_file(f'{filename}_000')
        # Test for invalid filename + context manager:
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.fileio',
                                 level='CRITICAL'):
                with FileIO('/#"½%&?<><|*', 'r', None) as some_file:
                    self.assertEqual(some_file.file_mode, 'r')
                    lines = list(some_file)
                    self.assertFalse(lines)
        # Test for weird mode:
        fileio = FileIO('some-file', 'q', None)
        with self.assertRaises(ValueError):
            fileio.open()
        fileio = FileIO('some-file', 'r', None)
        fileio.file_mode = 'q'
        with self.assertRaises(ValueError):
            fileio.open_file_read()
        fileio = FileIO('some-file', 'w', None)
        fileio.file_mode = 'q'
        with self.assertRaises(ValueError):
            fileio.open_file_write()

    def test_write_modes(self):
        """Test the write modes and backup settings."""
        with tempfile.NamedTemporaryFile() as tmp:
            with FileIO(tmp.name, 'w', None, backup=True) as fileio:
                self.assertIsNotNone(fileio.fileh)
                self.assertFalse(fileio.fileh.closed)
                fileio.write('some text')
            with FileIO(tmp.name, 'a', None, backup=True) as fileio:
                self.assertIsNotNone(fileio.fileh)
                self.assertFalse(fileio.fileh.closed)
                status = fileio.write('some text')
                self.assertTrue(status)
                status = fileio.write('some text', end=None)
                self.assertTrue(status)
                status = fileio.write(None)
                self.assertFalse(status)
            fileio = FileIO(tmp.name, 'w', None, backup=True)
            fileio.open()
            fileio.close()
            with turn_on_logging():
                with self.assertLogs('pyretis.inout.fileio',
                                     level='WARNING'):
                    fileio.write('text')
            fileio = FileIO(tmp.name, 'a', None, backup='True')
            with turn_on_logging():
                with self.assertLogs('pyretis.inout.fileio',
                                     level='WARNING'):
                    fileio.write('text')

    def test_reading(self):
        """Test generic reading."""
        filename = os.path.join(HERE, 'energy.txt')
        correct = []
        with open(filename, 'r', encoding='utf-8') as infile:
            correct = list(infile)
        lines = []
        with FileIO(filename, 'r', OutputFormatter('test')) as fileio:
            for line in fileio:
                lines.append(line)
        self.assertListEqual(lines, correct)
        with FileIO(filename, 'r', OutputFormatter('test')) as fileio:
            fileio.close()
            lines = list(fileio)
            self.assertFalse(lines)

    def test_load(self):
        """Test generic load."""
        filename = os.path.join(HERE, 'energy.txt')
        with FileIO(filename, 'r', OutputFormatter('test')) as fileio:
            for block in fileio.load():
                self.assertEqual(
                    block['comment'],
                    [('#     Time      Potential        Kinetic'
                      '          Total    Temperature')],
                )

    def test_output(self):
        """Test generic output."""
        data = [1, 2.0, 3.123456]
        with tempfile.NamedTemporaryFile() as tmp:
            with FileIO(tmp.name, 'w', OutputFormatter('test'),
                        backup=False) as fileio:
                fileio.output(1, data)
                fileio.output(2, data)
            tmp.flush()
            with FileIO(tmp.name, 'r', OutputFormatter('test')) as fileio:
                for block in fileio.load():
                    for i, line in enumerate(block['data']):
                        raw = [i + 1] + data
                        self.assertListEqual(line, raw)
        with tempfile.NamedTemporaryFile() as tmp:
            fileio = FileIO(tmp.name, 'w', OutputFormatter(''), backup=False)
            fileio.open()
            fileio.output(1, data)
            fileio.flush()
            fileio.close()
            fileio.flush()

    def test_formatter_info(self):
        """Test the formatter info method."""
        fileio = FileIO('My-name-is-nobody', 'r', None)
        info = fileio.formatter_info()
        self.assertIsNone(info)
        formatter = OutputFormatter('Yippee ki-yay')
        fileio = FileIO('My-name-is-nobody', 'r', formatter)
        info = fileio.formatter_info()
        self.assertEqual(info, OutputFormatter)


class TestEnergyFile(unittest.TestCase):
    """Test the energy file i/o."""

    def test_energy_fileio(self):
        """Test reading and writing for energy files."""
        fields = ['vpot', 'ekin', 'etot', 'temp']
        with tempfile.NamedTemporaryFile() as tmp:
            raw_data = []
            with EnergyFile(tmp.name, 'w') as efile:
                for i in range(50):
                    rnd = rand(len(fields))
                    raw_data.append([i] + list(rnd))
                    data = dict(zip(fields, rnd))
                    efile.output(i, data)
            tmp.flush()
            raw_data = np.array(raw_data)
            with EnergyFile(tmp.name, 'r') as efile:
                for block in efile.load():
                    data = block['data']
                    for i, key in enumerate(['time'] + fields):
                        for num1, num2 in zip(raw_data[:, i], data[key]):
                            self.assertAlmostEqual(num1, num2, 6)

    def test_energy_file_read(self):
        """Test reading existing energy data."""
        filename = os.path.join(HERE, 'energy.txt')
        correct = {
            'time': [0, 1000, 2000],
            'vpot': [-0.165063, -0.806879, -0.760215],
            'ekin': [0.035000, 0.328694, 0.041118],
            'etot': [-0.130063, -0.478185, -0.719097],
            'temp': [0.070000, 0.657387, 0.082236],
        }
        with EnergyFile(filename, 'r') as efile:
            for block in efile.load():
                for key, val in correct.items():
                    self.assertTrue(np.allclose(val, block['data'][key]))
        # Reading a file with errors, this should give a warning:
        filename = os.path.join(HERE, 'energy-error.txt')
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.fileio',
                                 level='WARNING'):
                with EnergyFile(filename, 'r') as efile:
                    for _ in efile.load():
                        pass


class TestCrossFile(unittest.TestCase):
    """Test the cross file i/o."""

    def test_cross_fileio(self):
        """Test reading and writing for cross files."""
        with tempfile.NamedTemporaryFile() as tmp:
            raw_data = []
            with CrossFile(tmp.name, 'w') as cfile:
                prev = 0
                for i in range(50):
                    interf = randint(1, 10)
                    step = prev + randint(1, 1000)
                    prev = step
                    direction = 1 if randint(0, 2) == 1 else -1
                    raw_data.append((step, interf + 1, direction))
                    data = [(step, interf, ('-', '+', '+')[direction + 1])]
                    cfile.output(i, data)
            tmp.flush()
            with CrossFile(tmp.name, 'r') as cfile:
                for block in cfile.load():
                    data = block['data']
                    for i, j in zip(data, raw_data):
                        self.assertEqual(i, j)

    def test_cross_fileread(self):
        """Test reading from existing files."""
        filename1 = os.path.join(HERE, 'cross-error.txt')
        filename2 = os.path.join(HERE, 'cross.txt')
        data1 = None
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.fileio',
                                 level='WARNING'):
                with CrossFile(filename1, 'r') as cfile:
                    for block in cfile.load():
                        data1 = block['data']
        data2 = None
        with CrossFile(filename2, 'r') as cfile:
            for block in cfile.load():
                data2 = block['data']
        for i, line in enumerate(data2):
            if i < 10:
                self.assertEqual(line, data1[i])
            elif i == 10:
                pass
            else:
                self.assertEqual(line, data1[i-1])


class TestOrderFile(unittest.TestCase):
    """Test the order parameter file i/o."""

    def test_order_parameter_fileio(self):
        """Test writing and reading of order parameter files."""
        extra = 3
        with tempfile.NamedTemporaryFile() as tmp:
            raw_data = []
            with OrderFile(tmp.name, 'w') as ofile:
                for i in range(50):
                    rnd = rand(1 + extra)
                    raw_data.append([i] + list(rnd))
                    ofile.output(i, rnd)
            tmp.flush()
            with OrderFile(tmp.name, 'r') as ofile:
                for block in ofile.load():
                    data = block['data']
                    for i, j in zip(data, raw_data):
                        for num1, num2 in zip(i, j):
                            self.assertAlmostEqual(num1, num2, 6)

    def test_order_parameter_read(self):
        """Test reading of order parameter data from a file."""
        filename = os.path.join(HERE, 'order-data.txt')
        correct_data = [
            [[0, 0.1, -0.1], [1, 0.2, -0.2], [2, 0.3, -0.3],
             [3, 0.4, -0.4], [4, 0.5, -0.5]],
            [[0, 1.1, -1.1], [1, 1.2, -1.2], [2, 1.3, -1.3],
             [3, 1.4, -1.4], [4, 1.5, -1.5]],
            [[0, 2.1, -2.1], [1, 2.2, -2.2], [2, 2.3, -2.3],
             [3, 2.4, -2.4], [4, 2.5, -2.5], [5, 2.6, -2.6],
             [6, 2.7, -2.7], [7, 2.8, -2.8]],
            [[0, 3.1, -3.1], [1, 3.2, -3.2], [2, 3.3, -3.3]],
            [[0, 4.1, -4.1], [2, 4.3, -4.3], [3, 4.4, -4.4]],
        ]
        with OrderFile(filename, 'r') as ofile:
            for i, data in enumerate(ofile.load()):
                self.assertEqual(
                    data['comment'][0],
                    f'# Cycle: {i}, status: ACC'
                )
                self.assertEqual(
                    data['comment'][1],
                    '#     Time       Orderp',
                )
                for j, k in zip(correct_data[i], data['data']):
                    self.assertTrue(np.allclose(j, k))


class TestPathFiles(unittest.TestCase):
    """Test file i/o for path files."""

    def test_energy_path_file(self):
        """Test i/o for energy path data."""
        paths, correct_data = create_test_paths(npath=5)
        with tempfile.NamedTemporaryFile() as tmp:
            for i, path in enumerate(paths):
                with EnergyPathFile(tmp.name, 'a') as efile:
                    efile.output(i, [path, 'ACC'])
            tmp.flush()
            with EnergyPathFile(tmp.name, 'r') as efile:
                for i, block in enumerate(efile.load()):
                    for key in ('ekin', 'vpot'):
                        self.assertTrue(np.allclose(block['data'][key],
                                                    correct_data[i][key]))

    def test_order_path_file(self):
        """Test i/o for order path data."""
        paths, correct_data = create_test_paths(npath=5)
        with tempfile.NamedTemporaryFile() as tmp:
            for i, path in enumerate(paths):
                with OrderPathFile(tmp.name, 'a') as ofile:
                    ofile.output(i, [path, 'ACC'])
            tmp.flush()
            with OrderPathFile(tmp.name, 'r') as ofile:
                for i, block in enumerate(ofile.load()):
                    corr = np.array(correct_data[i]['order'])
                    data = block['data']
                    self.assertTrue(np.allclose(data[:, 1], corr[:, 0]))
                    self.assertTrue(np.allclose(data[:, 2], corr[:, 1]))
                    self.assertTrue(np.allclose(data[:, 3], corr[:, 2]))

    def test_path_int_file(self):
        """Test the internal path writer."""
        paths, _ = create_test_paths(npath=3)
        with tempfile.NamedTemporaryFile() as tmp:
            for i, path in enumerate(paths):
                with PathIntFile(tmp.name, 'a') as pfile:
                    pfile.output(i, [path, 'ACC'])
            tmp.flush()
            with PathIntFile(tmp.name, 'r') as pfile:
                for i, block in enumerate(pfile.load()):
                    self.assertEqual(f'# Cycle: {i}, status: ACC',
                                     block['comment'][0])

    def test_path_ext_file(self):
        """Test the external path writer."""
        with tempfile.NamedTemporaryFile() as tmp:
            for i in range(3):
                path, _ = create_external_path(random_length=True)
                with PathExtFile(tmp.name, 'a') as pfile:
                    pfile.output(i, [path, 'ACC'])
            tmp.flush()
            with PathExtFile(tmp.name, 'r') as pfile:
                for i, block in enumerate(pfile.load()):
                    self.assertEqual(f'# Cycle: {i}, status: ACC',
                                     block['comment'][0])


class TestPathEnsembleFile(unittest.TestCase):
    """Test the i/o for the PathEnsembleFile."""

    @staticmethod
    def test_path_ensemble_read():
        """Test reading of path ensemble files."""
        filename = os.path.join(HERE, 'pathensemble001.txt')
        with PathEnsembleFile(filename, 'r') as pfile:
            for i, path in enumerate(pfile.get_paths()):
                ref_data = PATH_DATA[i]
                ref_data['filename'] = filename
                assert_equal_path_dict(path, ref_data)

    @staticmethod
    def _fake_path_from_dict(path_dict):
        """Return a path object from a path dict."""
        path = Path(None)
        path.generated = path_dict['generated']
        avg = 0.5 * (path_dict['ordermax'][0] + path_dict['ordermin'][0])
        for i in range(path_dict['length']):
            path.append(
                set_up_system([avg, i], [0], [0], vpot=0, ekin=0)
            )
        if path_dict['interface'][0] == 'L':
            path.phasepoints[0].order[0] = path_dict['ordermin'][0] + 0.0001
        if path_dict['interface'][0] == 'R':
            path.phasepoints[0].order[0] = path_dict['ordermax'][0]
        if path_dict['interface'][-1] == 'L':
            path.phasepoints[-1].order[0] = path_dict['ordermin'][0] + 0.0001
        if path_dict['interface'][-1] == 'R':
            path.phasepoints[-1].order[0] = path_dict['ordermax'][0]
        for key in ('ordermax', 'ordermin'):
            idx = path_dict[key][1]
            path.phasepoints[idx].order[0] = path_dict[key][0]
        return path

    def test_path_ensemble_write(self):
        """Test writing of path ensemble data."""
        filename = os.path.join(HERE, 'pathensemble001.txt')
        settings = {
            'ensemble_number': 1,
            'interfaces': [-0.9, -0.9, 1],
            'detect': -0.8
        }
        ens = PathEnsemble(settings['ensemble_number'],
                           settings['interfaces'])
        with tempfile.NamedTemporaryFile() as tmp:
            with PathEnsembleFile(tmp.name, 'w',
                                  ensemble_settings=settings) as pfile:
                for pathi in PATH_DATA:
                    path = self._fake_path_from_dict(pathi)
                    ens.add_path_data(path, pathi['status'],
                                      cycle=pathi['cycle'])
                    pfile.output(pathi['cycle'], ens)
            tmp.flush()
            self.assertTrue(filecmp.cmp(tmp.name, filename))
        # Test warning when ensemble settings are not given:
        with turn_on_logging():
            with self.assertLogs('pyretis.inout.formats.pathensemble',
                                 level='WARNING'):
                with PathEnsembleFile(filename, 'r') as pfile:
                    pass
        # Test open missing file:
        with PathEnsembleFile('path_mack_path/file__', 'r',
                              ensemble_settings=settings) as pfile:
            with turn_on_logging():
                with self.assertLogs('pyretis.inout.formats.pathensemble',
                                     level='CRITICAL'):
                    for _ in pfile.load():
                        pass


class TestSnapshot(unittest.TestCase):
    """Test methods related to the snapshot files."""

    def test_initiate_settings(self):
        """Test initiation of the snapshot file with settings."""
        settings = {'write_vel': False, 'extra-to-ignore': False}
        with tempfile.NamedTemporaryFile() as tmp:
            with SnapshotFile(tmp.name, 'r',
                              format_settings=settings) as sfile:
                self.assertFalse(sfile.formatter.write_vel)

    def test_read_txt_snapshot(self):
        """Test the read_txt_snapshot method."""
        filename = os.path.join(HERE, 'config.txt')
        read1 = read_txt_snapshots(filename)
        read2 = read_txt_snapshots(filename,
                                   data_keys=('name', 'x', 'y', 'z'))
        correct = {
            'atomname': [['X', 'Y'], ['A', 'B']],
            'box': [np.array([1., 2., 3., 4., 5., 6.]),
                    np.array([7., 8., 9.])],
            'x': [[1.0, 7.0], [1.1, 7.1]],
            'y': [[2.0, 8.0], [2.1, 8.1]],
            'z': [[3.0, 9.0], [3.1, 9.1]],
            'vx': [[4.0, 1.0], [4.1, 1.1]],
            'vy': [[5.0, 2.0], [5.1, 2.1]],
            'vz': [[6.0, 3.0], [6.1, 3.1]],
        }
        for i, (snap1, snap2) in enumerate(zip(read1, read2)):
            self.assertTrue(snap1['atomname'] == correct['atomname'][i])
            self.assertTrue(snap2['name'] == correct['atomname'][i])
            for j in ('x', 'y', 'z', 'box'):
                self.assertTrue(np.allclose(snap1[j], correct[j][i]))
                self.assertTrue(np.allclose(snap2[j], correct[j][i]))
            for j in ('vx', 'vy', 'vz'):
                self.assertTrue(np.allclose(snap1[j], correct[j][i]))
                self.assertFalse(j in snap2)
        filename = os.path.join(HERE, 'config_with_error.txt')
        read3 = read_txt_snapshots(filename)
        with self.assertRaises(Exception) as context:
            for i in read3:
                print(i)
        self.assertTrue('Oops_not_an_integer' in str(context.exception))
        self.assertTrue('invalid literal for int()' in str(context.exception))
        # Also test the file writer:
        filename = os.path.join(HERE, 'config.txt')
        with SnapshotFile(filename, 'r') as traj:
            for i, data in enumerate(traj.load()):
                for j in ('x', 'y', 'z', 'box'):
                    self.assertTrue(np.allclose(data[j], correct[j][i]))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains a WCA double well potential.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

DoubleWellWCA (:py:class:`.DoubleWellWCA`)
    This class defines a double well WCA potential.
"""
import logging
import numpy as np
from pyretis.forcefield.potential import PotentialFunction
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['DoubleWellWCA']


class DoubleWellWCA(PotentialFunction):
    r"""A double well potential.

    This class defines a double well WCA potential. The potential energy
    (:math:`V_\text{pot}`) for a pair of particles separated by a
    distance :math:`r` is given by,

    .. math::

       V_\text{pot} = h (1 - (r - r_0 - w)^2/w^2)^2,

    where :math:`h` gives the 'height' of the potential, :math:`r_0` the
    minimum and :math:`w` the width. These parameters are stored in the
    attributes `height`, `rzero` and `width` respectively.

    Attributes
    ----------
    params : dict
        Contains the parameters. These are:

        * `height`: A float describing the "height" of the potential.

        * `height4`: A float equal to ``4.0 * height``.
          (This variable is just included for convenience).

        * `rzero`: A float defining the two minimums. One is located at
          ``rzero``, the other at ``rzero+2*width``.

        * `types`: A set defining what kind of particle pairs to
          consider for this interaction. If `types` is not set (i.e.
          equal to None), it will be assumed to apply to **ALL**
          particles.

        * `width`: A float describing the "width" of the potential.

        * `width2`: A float equal to ``width*width`` (for convenience).

    """

    def __init__(self, dim=3, desc='A WCA double well potential'):
        """Initialise the Double Well WCA potential.

        Parameters
        ----------
        dim : int, optional
            The dimensionality of the potential.
        desc : string, optional
            Description of the force field.

        """
        super().__init__(dim=dim, desc=desc)
        self.params = {'height': 0.0,
                       'height4': 0.0,
                       'rwidth': 0.0,
                       'rzero': 0.0,
                       'types': [],
                       'width': 0.0,
                       'width2': 0.0}

    def set_parameters(self, parameters):
        """Add new potential parameters to the potential.

        Parameters
        ----------
        parameters : dict
            The new parameters, they are assume to be dicts on the form
            ``{'types': set([(0,0)]), 'rzero': 1.0, 'width': 0.25,
            'height': 6.0}``

        """
        for key in parameters:
            if key in self.params:
                self.params[key] = parameters[key]
            else:
                msg = f'Unknown parameter {key} - ignored!'
                logger.warning(msg)
        if self.params['types'] is not None:
            self.params['types'] = set(self.params['types'])
        self.params['width2'] = self.params['width']**2
        self.params['rwidth'] = self.params['rzero'] + self.params['width']
        self.params['height4'] = 4.0 * self.params['height']

    def _activate(self, itype, jtype):
        """Determine if we should calculate a interaction or not.

        Parameters
        ----------
        itype : string
            Particle type for particle i.
        jtype : string
            Particle type for particle j.

        """
        if self.params['types'] is None:
            return True
        pair1, pair2 = (itype, jtype), (jtype, itype)
        return (pair1 in self.params['types'] or
                pair2 in self.params['types'])

    def min_max(self):
        """Return the minima & maximum of the `DoubleWellWCA` potential.

        The minima are located at ``rzero`` & ``rzero + 2*width``.
        The maximum is located at ``rzero + width``.

        Returns
        -------
        out[0] : float
            Minimum number one, located at: ``rzero``.
        out[1] : float
            Minimum number two, located at: ``rzero + 2*width``.
        out[2] : float
            Maximum, located at: ``rzero + width``.

        """
        rzero = self.params['rzero']
        width = self.params['width']
        return rzero, rzero + 2.0 * width, rzero + width

    def potential(self, system):
        """Calculate the potential energy.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential in.

        Returns
        -------
        v_pot : float
            The potential energy.

        """
        particles = system.particles
        box = system.box
        v_pot = 0.0
        rwidth = self.params['rwidth']
        width2 = self.params['width2']
        height = self.params['height']
        for pair in particles.pairs():
            i, j, itype, jtype = pair
            if self._activate(itype, jtype):
                delta = box.pbc_dist_coordinate(particles.pos[i] -
                                                particles.pos[j])
                delr = np.sqrt(np.dot(delta, delta))
                v_pot += (height * (1.0 - (((delr - rwidth)**2) / width2))**2)
        return v_pot

    def force(self, system):
        """Calculate the force.

        We also calculate the virial here, since the force is evaluated.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential in.

        Returns
        -------
        forces : numpy.array
            The force as a numpy.array of the same shape as the
            positions in `particles.pos`.
        virial : numpy.array
            The virial, as a symmetric matrix with dimensions (dim, dim)
            where dim is given by the box.

        """
        particles = system.particles
        box = system.box
        forces = np.zeros(particles.pos.shape)
        virial = np.zeros((box.dim, box.dim))
        for pair in particles.pairs():
            i, j, itype, jtype = pair
            if self._activate(itype, jtype):
                delta = box.pbc_dist_coordinate(particles.pos[i] -
                                                particles.pos[j])
                delr = np.sqrt(np.dot(delta, delta))
                diff = delr - self.params['rwidth']
                forceij = (
                    self.params['height4'] *
                    (1.0 - diff**2 / self.params['width2']) *
                    (diff / self.params['width2'])
                )
                forceij = forceij * delta / delr
                forces[j] -= forceij
                forces[i] += forceij
                virial += np.outer(forceij, delta)
        return forces, virial

    def potential_and_force(self, system):
        """Calculate the force & potential.

        We also calculate the virial here, since the force is evaluated.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential in.

        Returns
        -------
        out[0] : float
            The potential energy as a float.
        out[1] : numpy.array
            The force as a numpy.array of the same shape as the
            positions in `particles.pos`.
        out[2] : numpy.array
            The virial, as a symmetric matrix with dimensions (dim, dim)
            where dim is given by the box.

        """
        particles = system.particles
        forces = np.zeros(particles.pos.shape)
        virial = np.zeros((system.box.dim, system.box.dim))
        v_pot = 0.0
        for pair in particles.pairs():
            i, j, itype, jtype = pair
            if self._activate(itype, jtype):
                delta = system.box.pbc_dist_coordinate(particles.pos[i] -
                                                       particles.pos[j])
                delr = np.sqrt(np.dot(delta, delta))
                diff = delr - self.params['rwidth']
                v_pot += (self.params['height'] *
                          (1.0 - diff**2 / self.params['width2'])**2)
                forceij = (self.params['height4'] *
                           (1.0 - diff**2 / self.params['width2']) *
                           (diff / self.params['width2']))
                forceij = forceij * delta / delr
                forces[i] += forceij
                forces[j] -= forceij
                virial += np.outer(forceij, delta)
        return v_pot, forces, virial



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This package contains methods for generating reports.

The reports will typically summarise the results from different
analysis and present it as a text file, pdf or web-page.

Package structure
-----------------

Modules
~~~~~~~

__init__.py
    This file. Handles imports for PyRETIS and defines a method for
    writing a report to a file.

markup.py (:py:mod:`pyretis.inout.report.markup`)
    This module defines some methods for generating simple tables and
    formatting numbers for rst and latex.

report.py (:py:mod:`pyretis.inout.report.report`)
    General methods for generating reports.

report_md.py (:py:mod:`pyretis.inout.report.report_md`)
    This module defines the molecular dynamics reports. Specifically
    it defines the report that is made based on results from a MD Flux
    simulations.

report_path.py (:py:mod:`pyretis.inout.report.report_path`)
    This module defines the reports for path simulations like TIS and
    RETIS.


Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

generate_report (:py:func:`.generate_report`)
    Method for generating reports.

Folders
~~~~~~~

templates
    A folder containing templates for generating reports.
"""
from .report import generate_report
